#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <time.h>
#include <sys/stat.h>
#include <errno.h>

#ifndef WIN32 
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#else
#include <winsock.h>
#endif

#include "config.h"
#include "socket.h"
#include "login.h" 
#include "data.h"

extern unsigned int mode_passiveport;
extern unsigned int mode_usleep;


void data_setfd_set(fd_set *r, fd_set *w, struct data_node *d)
{

  for (; d; d=d->next) {

    if (d->status & ST_WRITE) {

      FD_SET(d->socket, w);
      continue;

    }

    if (d->status & ST_WRITE2) {

      FD_SET(d->altsocket, w);
      continue;

    }


    if (d->status & ST_LISTEN) {  /* We're waiting for a connect to shadow */

      if (d->type == DATA_PORT)
		FD_SET(d->altsocket, r);
      else {
		FD_SET(d->socket, r);
	  }
    } 

    else if (d->status & ST_PENDING) { /* we're connecting somewhere */

      if (d->type == DATA_PORT) {
	FD_SET(d->socket, r);
	FD_SET(d->socket, w);
      } else {
	FD_SET(d->altsocket, r);
	FD_SET(d->altsocket, w);
      }
    }

    if (d->status & ST_REMOTE) {

      FD_SET(d->altsocket, r);

    }

    if (d->status & ST_ACTIVE) {

      FD_SET(d->socket, r);

    }

  }

}


void data_input(fd_set *r, fd_set *w, struct data_node *d)
{

  for ( ; d; d = d->next) {

    if (d->status & ST_WRITE) {

      if (FD_ISSET(d->socket, w)) {

		if (!data_get(d, d->altsocket, d->socket, 0)) return;

      }

      continue;

    }


    if (d->status & ST_WRITE2) {

      if (FD_ISSET(d->altsocket, w)) {

		if (!data_get(d, d->socket, d->altsocket, 0)) return;

      }

      continue;

    }





    if (d->status & ST_LISTEN) {

      if ((d->type == DATA_PORT) && FD_ISSET(d->altsocket, r)) {

	/* Connection successful */
	
	unsigned long host;
	int sock, port;

	printf("Calling accept\n");

	sock = socket_accept(d->altsocket, &host, &port);

	if (sock == -1) { /* Bah, it failed. */
	  socket_print(d->login->socket, "425 Can't build data connection(4).\r\n");
	  close(d->altsocket);
	  d->status |= ST_LOGOUT;
	  continue;
	}

	printf("after accept(): closing %d\n", d->altsocket);

	close(d->altsocket);
	d->altsocket = sock;
	
	printf("Incomming connection established(%d). %s:%d.\n", sock,
	       ul2a(host), port);

	d->status &= ~ST_LISTEN;
	d->status |= ST_REMOTE;

	printf("Connecting to: %s %d\n", ul2a(d->port_host),d->port_port);

	/* Fire off out own connection to user's PORT */
	d->socket = socket_connect(d->port_host, d->port_port);

	printf("Connect returned %d %d\n", d->socket, errno);

	if (d->socket < 0) { /* bah, failed to connect to them */
	  socket_print(d->login->socket, "425 Can't build data connection(5).\r\n");
	  d->status |= ST_LOGOUT;
	  continue;
	}

	printf("Pending\n");
	d->status |= ST_PENDING; /* Tell us we're awaiting a connection */

	continue;

      }

      else if ((d->type == DATA_PASSIVE) && FD_ISSET(d->socket, r)) {
	
	/* Connection successful */
	
	unsigned long host;
	int sock, port;
	
	printf("Calling accept (passive)\n");
	
	sock = socket_accept(d->socket, &host, &port);
	
	if (sock == -1) { /* Bah, it failed. */
	  socket_print(d->login->socket, "425 Can't build data connection(6).\r\n");
	  close(d->socket);
	  d->status |= ST_LOGOUT;
	  continue;
	}

	printf("After accept closing %d\n", d->socket);
	close(d->socket);
	d->socket = sock;

	printf("Incomming connection established(%d). %s:%d.\n", sock,
	       ul2a(host), port);

	d->status &= ~ST_LISTEN;
	d->status |= ST_ACTIVE;

	printf("Connecting to: %s %d\n", ul2a(d->shadow_host),d->shadow_port);

	/* Fire off out own connection to user's PORT */
	d->altsocket = socket_connect(d->shadow_host, d->shadow_port);

	if (d->altsocket < 0) { /* bah, failed to connect to them */
	  socket_print(d->login->socket, "425 Can't build data connection(7).\r\n");
	  d->status |= ST_LOGOUT;
	  continue;
	}

	printf("Pending\n");
	d->status |= ST_PENDING; /* Tell us we're awaiting a connection */

	continue;

      }

      continue;

    }


    if (d->status & ST_PENDING) { /* check our connection */

      if (d->type == DATA_PORT) {

	if (FD_ISSET(d->socket, w)) { /* connection succeeded */
	  
	  d->status &= ~ST_PENDING;
	  d->status |= ST_ACTIVE;
	  
	  printf("Connected\n");
	  
	} else if (FD_ISSET(d->socket, r)) { /* Connection failed */
	  
	  d->status |= ST_LOGOUT;
	  printf("Failed\n");
	  
	}
      }

      else if (d->type == DATA_PASSIVE) {

	if (FD_ISSET(d->altsocket, w)) { /* connection succeeded */
	  
	  d->status &= ~ST_PENDING;
	  d->status |= ST_REMOTE;
	  
	} else if (FD_ISSET(d->altsocket, r)) { /* Connection failed */

	  d->status |= ST_LOGOUT;
	  printf("Failed2\n");
	  
	}
	
      }

      continue;

    }
    

    if ((d->status & ST_ACTIVE) && (d->status & ST_REMOTE)) {

      if (FD_ISSET(d->socket, r)) { /* input from user */
	
		if (!data_get(d, d->socket, d->altsocket, 0)) return;
	
      } 
      
      if ((d->altsocket >= 0) && (d->altsocket <= 256))
	if (FD_ISSET(d->altsocket, r)) { /* input from user */
	
	  if (!data_get(d, d->altsocket, d->socket, 0)) return;
	
      } 
    }
  }
}



int data_get(struct data_node *d, int in, int out, int mode)
{
  static unsigned char buffer[DATA_BUFSZ];
  unsigned char *buf_ptr = NULL;
  int red, wrote;
  


  /* Not a POSIX call */
#if !defined ( LINUX ) && !defined (WIN32)
  if (mode_usleep)
    usleep(mode_usleep);
#endif



  if (!d->login->outbuffer) {
    
    buf_ptr = buffer;
    
    if ((red = read(in, buffer, DATA_BUFSZ)) <= 0) {
      
      d->status |= ST_LOGOUT;
      
      data_close(d); /* Shouldn't be required :( */

      
      return 0;

    } /* read() */
    

  } else {
    /* If we have something in the outbuffer already, skip read and use the 
       buffer */
    
    buf_ptr = d->login->obuffer;
    red = d->login->outbuffer;
    
    d->login->outbuffer = 0;
    
  }

  /* 
   * we now have DATA if but_ptr, and red is size 
   */

  if ((wrote = write(out, buf_ptr, red)) != red) {
    
    if (wrote <= 0) {

#ifndef WIN32
      if ((errno == EAGAIN) || (errno == EWOULDBLOCK)) { /* nonblock pause */
#else
      if ((errno == EAGAIN) || (errno == WSAEWOULDBLOCK)) { /* nonblock pause */
#endif

	
	wrote = 0; /* We wrote nothing - copy all to back buffer */

#if 0
	bcopy(&buf_ptr[0], d->login->obuffer, red);
	d->login->outbuffer = red;
	return 0;
#endif

      } else {
	
	perror("Bah");
	printf("File closed on write %d %d\n", red, wrote);
	
	d->status |= ST_LOGOUT;
	return 0; /* Socket closed */

      }
      
    }
    
    /* We wrote 'red', but only got wrote byte out. */
    bcopy(&buf_ptr[wrote], d->login->obuffer, red - wrote);
    d->login->outbuffer = red - wrote;

    if (out == d->socket)
      d->status |= ST_WRITE;
    else
      d->status |= ST_WRITE2;
  }


  d->status &= ~ST_WRITE;
  d->status &= ~ST_WRITE2;

  return 1;

}



#if 0
void data_close(struct data_node *d)
{
  struct data_node *curr, *last = NULL;
  struct quota_node *quota;
  time_t tt;
  
  d->login->data_active--;
  
  for (curr = (struct data_node *)d->login->data; 
       curr; last = curr, curr=curr->next) {
    
    if (curr == d) {
      if (!last)
        d->login->data = (void *) curr->next;
      else
        last->next = curr->next;
      
      free(d);
      break;
      
    }
  }
}
#endif

struct data_node *data_new()
{
  struct data_node *d;

  d = (struct data_node *) malloc(sizeof(struct data_node));

  memset(d, 0, sizeof(struct data_node));

  d->socket = -1;
  d->altsocket = -1;
  d->status = 0;

  return d;
}



int data_port(struct login_node *t, unsigned char *buffer, int size,
	      int in, int out)
{
  struct data_node *d;
  unsigned char *pp;
  unsigned char *pp2;
  int IPA[4], PORTA[2], pickone = 0, len;
  int IPA2[4], PORTA2[2];
  struct sockaddr_in par;


  printf("Intercepting PORT command:\n");

  buffer[size] = 0;

  /* 
   * Flow:
   * Create new data node
   * - set data node's host/port to PORT strings.
   * - open a new socket for our PORT
   * send fake PORT to shadowftpd.
   */

  if (!(d = data_new())) {
    socket_print(in, "425 Can't build data connection(1).\r\n");
    return 0;
  }

  d->type = DATA_PORT;
  d->login = t;

  if (sscanf(&buffer[5], "%d,%d,%d,%d,%d,%d", &IPA[0], &IPA[1], &IPA[2], &IPA[3],
	     &PORTA[0], &PORTA[1]) != 6) {
    socket_print(in, "425 Can't build data connection(2).\r\n");
    free(d);
    return 0;
  }

  pp = (unsigned char *)&(d->port_host);
  pp[0] = (unsigned char) IPA[0];
  pp[1] = (unsigned char) IPA[1];
  pp[2] = (unsigned char) IPA[2];
  pp[3] = (unsigned char) IPA[3];

  pp = (unsigned char *)&(d->port_port);
  pp[0] = (unsigned char) PORTA[0];
  pp[1] = (unsigned char) PORTA[1];

  d->port_port = ntohs(d->port_port);

  printf("data_port (%d): %s:%d\n",  t->socket, ul2a(d->port_host),
         d->port_port);
  
  
  /* Right, now make a new socket for shadow FTPD */
  
  if ((d->altsocket = socket_make(&pickone)) == -1) {
    /* Can't Passive */
    socket_print(in, "425 Can't build data connection(3).\r\n");
    free(d);
    return 0;
  }

  /* 227 Entering Passive Mode (1,1,1,6,4,8) */

  len = sizeof(par);
  
  if (getsockname(t->altsocket, 
                  (struct sockaddr *)&par, &len) < 0) {
    
    bcopy(&localhost, &(d->shadow_host), sizeof(localhost));

  } else
    bcopy(&par.sin_addr, &(d->shadow_host), sizeof(d->shadow_host));

  pp2 = (unsigned char *)&(d->shadow_host);
  IPA2[0] = (unsigned char) pp2[0];
  IPA2[1] = (unsigned char) pp2[1];
  IPA2[2] = (unsigned char) pp2[2];
  IPA2[3] = (unsigned char) pp2[3];

  d->shadow_port = htons(pickone);

  pp2 = (unsigned char *)&(d->shadow_port);
  PORTA2[0] = pp2[0];
  PORTA2[1] = pp2[1];
  
  socket_print(out, "PORT %d,%d,%d,%d,%d,%d\r\n", 
	       IPA2[0], IPA2[1], IPA2[2], IPA2[3], PORTA2[0], PORTA2[1]);

  printf("PORT %d,%d,%d,%d,%d,%d\r\n", 
	 IPA2[0], IPA2[1], IPA2[2], IPA2[3], PORTA2[0], PORTA2[1]);
  

  d->next = (struct data_node *)t->data;
  t->data = (void *) d; 

  d->status |= ST_LISTEN;
  printf("%d is ready to accept()\n", d->altsocket);

  printf("portport %d, shadowport %d\n", d->port_port, ntohs(d->shadow_port));

  return 1;
}


int data_pasv(struct login_node *t, unsigned char *buffer, int size,
	      int in, int out)
{

  /* We've just received a passive reply from someone
   * Fetch out the host and port
   * Open our own socket and send that passive reply instead
   * when we get a connect, connect back to passive host 
   */

  struct data_node *d;
  unsigned char *pp;
  int IPA[4], PORTA[2], pickone = 0, len;
  struct sockaddr_in par;


  printf("Intercepting PASV reply\n");

  /* 
   * Flow:
   * Create new data node
   * - set data node's host/port to PORT strings.
   * - open a new socket for our PORT
   * send fake PORT to shadowftpd.
   */

  if (!(d = data_new())) {
    socket_print(out, "425 Can't build data connection.\r\n");
    return 0;
  }

  d->type = DATA_PASSIVE;
  d->login = t;

  pp = strchr(buffer, '(');

  if (!pp) {
    socket_print(out, "425 Can't build data connection.\r\n");
    free(d);
    return 0;
  }


  if (sscanf(&pp[1], "%d,%d,%d,%d,%d,%d", &IPA[0], &IPA[1], &IPA[2], &IPA[3],
	     &PORTA[0], &PORTA[1]) != 6) {
    socket_print(out, "425 Can't build data connection.\r\n");
    free(d);
    return 0;
  }

  pp = (unsigned char *)&(d->shadow_host);
  pp[0] = (unsigned char) IPA[0];
  pp[1] = (unsigned char) IPA[1];
  pp[2] = (unsigned char) IPA[2];
  pp[3] = (unsigned char) IPA[3];

  pp = (unsigned char *)&(d->shadow_port);
  pp[0] = (unsigned char) PORTA[0];
  pp[1] = (unsigned char) PORTA[1];

  d->shadow_port = ntohs(d->shadow_port);

  printf("data_pasv (%d): %s:%d\n",  t->socket, ul2a(d->shadow_host),
         d->shadow_port);
  
  
  /* Right, now make a new socket for shadow FTPD */
  
  pickone = mode_passiveport;
  if ((d->socket = socket_make(&pickone)) == -1) {
    /* Can't Passive */
    socket_print(out, "425 Can't build data connection.\r\n");
    free(d);
    return 0;
  }

  /* 227 Entering Passive Mode (1,1,1,6,4,8) */

  len = sizeof(par);
  
#ifdef WIN32
  if ((getsockname(t->socket, 
                  (struct sockaddr *)&par, &len) == SOCKET_ERROR) || (par.sin_addr.s_addr == 0)) {
#else
  if (getsockname(t->socket, 
                  (struct sockaddr *)&par, &len) < 0) {
#endif

    
	printf("Copying localhost\n");
    bcopy(&localhost, &(d->port_host), sizeof(localhost));

  } else {

	printf("Not localhost\n");
    bcopy(&par.sin_addr, &(d->port_host), sizeof(d->port_host));
  }

  pp = (unsigned char *)&(d->port_host);
  IPA[0] = (unsigned char) pp[0];
  IPA[1] = (unsigned char) pp[1];
  IPA[2] = (unsigned char) pp[2];
  IPA[3] = (unsigned char) pp[3];

  d->port_port = htons(pickone);

  pp = (unsigned char *)&(d->port_port);
  PORTA[0] = pp[0];
  PORTA[1] = pp[1];
  
  socket_print(out, "227 Entering Passive Mode (%d,%d,%d,%d,%d,%d)\r\n", 
	       IPA[0], IPA[1], IPA[2], IPA[3], PORTA[0], PORTA[1]);

  printf("227 Entering Passive Mode (%d,%d,%d,%d,%d,%d)\r\n", 
	 IPA[0], IPA[1], IPA[2], IPA[3], PORTA[0], PORTA[1]);
  

  d->next = (struct data_node *)t->data;
  t->data = (void *) d; 

  d->status |= ST_LISTEN;
  printf("%d is ready to accept()\n", d->socket);

  return 1;
}



void data_close(struct data_node *d) 
{
  struct data_node **prev, *runner;

  printf("Deleting %d %d\n", d->socket, d->altsocket);

  d->login->data_active--;

  if (d->socket >= 0) {
    close(d->socket);
    d->socket = -1;
  }

  if (d->altsocket >= 0) {
    close(d->altsocket);
    d->altsocket = -1;
  }


  for (runner = (struct data_node *)d->login->data, 
       prev = (struct data_node **)&d->login->data;
       runner; prev = &runner->next, runner = runner->next) {

    if (d == runner) { /* delete this one, ace! */
      

      *prev = runner->next;

	  printf("Removing data node! %08X\n", d->login->data);

      free(runner);

      return;

    }

  }

}

void data_exit_check(struct data_node *runner)
{
  struct data_node *next;

  for (; runner; runner = next) {

    next = runner->next;

    if (runner->status & ST_LOGOUT) {

      printf("data flagged for delete\n");
      data_close(runner);

    }

  }      

}

