/*
 * Identd Prober (c) Jorgen Lundman
 *
 * Nonblocking identd lookup functions.
 *
 * Adds identd requests to an internal list, fires away requests and waits
 * for response or timeout.
 *
 * Shit, since when was this TCP :)
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include <signal.h>

#ifndef WIN32 

#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#else

#include <winsock.h>

#endif

#include <string.h>
#include <time.h>  /* linux */

#include <sys/types.h>

#include "config.h"
#include "socket.h"
#include "ident.h"

/* this technically belongs to .h file */
struct ident_node {

  int socket;

  int status;

  char *reply; /* where to put reply, set to 0 before request, filled in l8r */

  struct sockaddr_in ident;
  unsigned int local_port, remote_port;

  time_t time; /* for timeout */

  struct ident_node *next;
};


static struct ident_node *ident_node_head = 0;

static char ident_buf[256];


void login_ident_reply(char *);


void chop(char *s)
{
  char *r;

  if ((r=(char *) strchr(s, '\r')))
    *r = (char) 0;
  if ((r=(char *) strchr(s, '\n')))
    *r = (char) 0;
}



/*
 * Ready the identd prober port, structures etc. 
 */

void ident_init(void) 
{
}


void ident_request(int fd, char *reply)
{
  /* Ask for ident of socket fd */
  unsigned long host;
  struct sockaddr_in tmp;
  struct ident_node *ni;
  int len;

#ifdef WIN32
  strcpy(reply, FROM_UNKNOWN);
  return;
#endif


  ni = (struct ident_node *) malloc(sizeof(struct ident_node));

  if (!ni) {
    perror("ident_request(malloc())");
    strcpy(reply, FROM_UNKNOWN);
#if 0
    strcpy(reply, "NONE1");
#endif
    return;
  }
  
  memset(ni, 0, sizeof(*ni));

  ni->reply = reply;

  len = sizeof(struct sockaddr);

#if 1
  if(getpeername(fd, (struct sockaddr *)&ni->ident, &len) == -1) {
    /* error */
    perror("getpeername");
    strcpy(reply, FROM_UNKNOWN);
#if 0
    strcpy(reply, "NONE2");
#endif
    free(ni);
    return;
  }

  ni->local_port = ni->ident.sin_port;

#endif

  len = sizeof(struct sockaddr_in);

  if(getsockname(fd, (struct sockaddr *)&tmp,&len) == -1) {
    perror("getsockname");
    strcpy(reply, FROM_UNKNOWN);
#if 0
    strcpy(reply, "NONE2");
#endif
    free(ni);
    return;
  }

  ni->remote_port = tmp.sin_port;
  ni->ident.sin_port = ntohs(RFC931_PORT);
  
  printf("Firing off connection to %s:%d\n",
#ifdef WIN32
	   inet_ntoa(ni->ident.sin_addr),
#else
	   inet_ntoa((struct in_addr)ni->ident.sin_addr),
#endif
	   ntohs(ni->ident.sin_port));


  memcpy(&host, &ni->ident.sin_addr, sizeof(host));

  if ((ni->socket = socket_connect(host, ntohs(ni->ident.sin_port))) < 0) {

    perror("socket_connect()");
    free(ni);
    strcpy(reply, FROM_UNKNOWN);
#if 0
    strcpy(reply, "NONE3");
#endif
    return;

  }

  ni->reply[0] = 0; /* signal its being process */ 

  time(&ni->time); /* ready timeout */

  /* Ok, we're hoping for a connection */
  ni->status = IDENT_PENDING;

  /* add to linked list */
  ni->next = ident_node_head;
  ident_node_head = ni;

}


void ident_setfd_set(fd_set *r, fd_set *w)
{
  struct ident_node *in;

  for (in = ident_node_head; in; in = in->next) {

    if (in->status & IDENT_PENDING) {

      FD_SET(in->socket, r);
      FD_SET(in->socket, w);

    } else if (in->status & IDENT_ACTIVE) {

      FD_SET(in->socket, r);

    }    
  }
}


void ident_close(struct ident_node *in)
{
  /* ok, we have a reply from in, bad or good */

  printf("Closing identd %d\n", in->socket);
  close(in->socket);

  /* we call login.c so it can signal the login to continue, if it was
   * paused */

  login_ident_reply(in->reply);

  free(in);

}



void ident_input(fd_set *r, fd_set *w)
{
  unsigned int rmt_port, our_port;
  struct ident_node *in, *in_next, *prev;
  char reply[11];
  int i;

  for (in = ident_node_head, prev = NULL; in; prev = in, in = in_next) {
    
    in_next = in->next;
	  
    if (in->status & IDENT_PENDING) { /* connection pending */



if (FD_ISSET(in->socket, r)) { /* oops, no good */

	printf("Ident connection failed!\n");

	strcpy(in->reply, FROM_UNKNOWN);
#if 0
	strcpy(in->reply, "NONE4");
#endif
	/* remove node */
	if (!prev) 
	  ident_node_head = in_next;
	else
	  prev->next = in_next;
	
	ident_close(in);

	in = prev;
	continue;

      } else if (FD_ISSET(in->socket, w)) { /* connected */
	
	in->status &= ~IDENT_PENDING;
	in->status |= IDENT_ACTIVE;

	/* send request */
	sprintf(ident_buf, "%u,%u\r\n",
		ntohs(in->local_port),
		ntohs(in->remote_port));

	write(in->socket, ident_buf, strlen(ident_buf));

      }  

    } else if (in->status & IDENT_ACTIVE) {
      
      if (FD_ISSET(in->socket, r)) { /* we have something to read finally ! */
	
	/* We have a ident reply! */

	if ((i = read(in->socket, ident_buf, 256)) > 0) {
	  
	  ident_buf[i] = 0;

	  printf("Received ident reply from %s '%s'\n", 
#ifdef WIN32
			 inet_ntoa(in->ident.sin_addr),
#else
			 inet_ntoa((struct in_addr)in->ident.sin_addr),
#endif
		   ident_buf);
	  
	  /* find the machine node */
	  
	  /* Found 'em! */
	  if ((sscanf(ident_buf, "%u , %u : USERID :%*[^:]:%9s",
		      &rmt_port, &our_port, reply) == 3)  &&
	      (ntohs(in->remote_port) == our_port) &&
	      (ntohs(in->local_port) == rmt_port)) {

	    chop(reply);
	    
	    strcpy(in->reply, reply); /* signal we have found it! */
	    
	  } else {

	    strcpy(in->reply, FROM_UNKNOWN);
#if 0
	    strcpy(in->reply, "NONE5");
#endif
	  }
	  
	  /* remove node */
	  if (!prev) 
	    ident_node_head = in_next;
	  else
	    prev->next = in_next;
	  
	  ident_close(in);

	  in = prev;
	  continue;

	} /* read */
      } /* fdisset */
    } /* if active */
  } /* for */
}


void ident_timeout(void)
{
  struct ident_node *in, *in_next, *prev;
  time_t now;

  time(&now);

  for (in = ident_node_head, prev = NULL; in; in = in_next) {

#ifdef DEBUG
    printf("Pending ident-> %d %d %s (%d : %d) %u\n",
	     in->socket, in->status, 
#ifdef WIN32
	     inet_ntoa(in->ident.sin_addr),
#else
	     inet_ntoa((struct in_addr)in->ident.sin_addr),
#endif
	     ntohs(in->local_port), ntohs(in->remote_port),
	     in->time);
#endif
    
    in_next = in->next;
    
    if ((in->time + IDENT_TIMEOUT) < now) {
      
      printf("Timeout from %s %d\n", 
#ifdef WIN32
			 inet_ntoa(in->ident.sin_addr),
#else
			 inet_ntoa((struct in_addr)in->ident.sin_addr),
#endif
	       ntohs(in->local_port));
      
      strcpy(in->reply, FROM_UNKNOWN);
#if 0
      strcpy(in->reply, "NONE6");
#endif      
      /* remove node */
      if (!prev) 
	ident_node_head = in_next;
      else
	prev->next = in_next;
      
      ident_close(in);

    } else prev = in;  /* timeout */
  }
}


void ident_remove(char *reply)
{
  struct ident_node *in, *in_next, *prev;
  for (in = ident_node_head, prev = NULL; in; prev = in, in = in_next) {

    in_next = in->next;

    if (reply == in->reply) { /* remove us */

      if (!prev) 
	ident_node_head = in_next;
      else
	prev->next = in_next;
      
      close(in->socket);

      break;
    }
  }
}



