#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32 
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#else
#include <winsock.h>
#endif
#include <sys/stat.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>

#include "login.h"
#include "socket.h"
#include "config.h"
#include "data.h"
#include "ident.h"

struct login_node *login_node_head = NULL;


#if defined (SUNOS) || defined (IRIX)
extern int errno;
#endif

#if !defined ( LINUX ) && !defined ( IRIX )
#ifndef sys_errlist
extern const char *const sys_errlist[];
#endif
#endif


extern unsigned int mode_remoteport;
extern unsigned int mode_inetd;


void login_setfd_set(fd_set *r, fd_set *w)
{
  struct login_node *t;
  
  for (t=login_node_head; t; t=t->next) {

    FD_SET(t->socket, r);

    if (t->status & ST_REMOTE) {
      FD_SET(t->altsocket, r);
    }

    if (t->status & ST_PENDING) {
      FD_SET(t->altsocket, r);
      FD_SET(t->altsocket, w);
    }

    /* Add all data connections */
    if (t->data) 
      data_setfd_set(r, w, (struct data_node *)t->data);
    
  }
  
}




void login_input(fd_set *r, fd_set *w)
{
  struct login_node *t;

  for (t=login_node_head; t; t=t->next) {

    /* First check the control session */

    if (FD_ISSET(t->socket, r)) {
      /* There is data to read on t */

      /*             from       to  */
      login_get(t, t->socket, t->altsocket);

    }


    if (t->status & ST_REMOTE) {
      /* We already have a remote connection, check it for input */
      
      if (FD_ISSET(t->altsocket, r)) {
	/* there is incomming traffic from shadow host */

	/*             from       to  */
	login_get(t, t->altsocket, t->socket);

      }
    }


    if (t->status & ST_PENDING) {  /* waiting for a shadow connect */


	  printf("Testing for conn %d %d\n", FD_ISSET(t->altsocket, w), FD_ISSET(t->altsocket, r));


      if (FD_ISSET(t->altsocket, w)) { 
	/* We have a successful connect  */

	t->status &= ~ST_PENDING;
	t->status |= ST_REMOTE;


#if 1 /* experimental */
	if (mode_auth) {
	  printf("Sending: IDEN %s@%s\n", t->ident, ul2a(t->host));
	  
	  socket_print(t->altsocket, "IDEN %s@%s\r\n",
		       t->ident, ul2a(t->host));
	}
	
#else

	if (t->status & ST_IDENT) {

	  t->status &= ~ST_IDENT;

	  if (mode_auth) {
	    printf("Sending: IDEN %s@%s\n", t->ident, ul2a(t->host));

	    socket_print(t->altsocket, "IDEN %s@%s\r\n",
			 t->ident, ul2a(t->host));
	  }
	  
	}

#endif

	printf("Shadow connect complete: fd %d / %d\n",
	       t->socket, t->altsocket);

	//      } else {

	if (FD_ISSET(t->altsocket, r)) { 
	  /* connection to shadow failed */

	  printf("Shadow connect failed\n");
	  
	  t->status &= ~ST_PENDING;
	  t->status &= ~ST_REMOTE;

	  socket_print(t->socket, "599 ShadowFTPD connect failed.\r\n");
	  
	}
      }
    }	

    /* check data sessions */
    if (t->data)
      data_input(r, w, (struct data_node *)t->data);

  }
}



void login_new(int acc)
{
  int sock, port;
  unsigned long host;
  struct login_node *nue;


  if (!mode_inetd) {

    sock = socket_accept(acc, &host, &port);
    if (sock == -1)
      return;
    
    printf("New login(%d) from %s:%d\n", sock, ul2a(host), port);

  } else {
    
    sock = acc;

  }


  nue = (struct login_node *) malloc(sizeof(struct login_node));
  if (!nue || 
      !(nue->buffer = (unsigned char *) malloc(INPUT_BUFFER_SIZE)) ||
      !(nue->obuffer= (unsigned char *) malloc(DATA_BUFSZ))) {

    perror("Creating login");

    if (nue) {
      free(nue);

      if (nue->buffer)
        free(nue->buffer);
    }
    
    if (!mode_inetd) 
      close(sock);

    return ;

  }

  nue->status = ST_NONE;
  nue->socket = sock;
  nue->altsocket = -1;
  nue->inbuffer = 0;
  nue->outbuffer = 0;
  nue->host = host;
  nue->port = port;
  nue->data_active = 0;
  nue->data = 0;

  nue->next = login_node_head;
  login_node_head = nue;

  /* Fire away a identd lookup */
  if (!mode_inetd) 
    ident_request(sock, nue->ident);
  
  /* Fire away a connection to remote host */
  
  printf("Connecting to shadowhost...\n");

  nue->altsocket = socket_connect(shadowhost, mode_remoteport);

  if (nue->altsocket < 0) {
    socket_print(nue->socket, "599 %s\r\n", sys_errlist[errno]);
    login_exit(nue);
    return;
  }

  nue->status |= ST_PENDING;

}

void login_exit(struct login_node *t)
{
  struct login_node *curr, *last = NULL;
  /*  struct data_node *data;*/


  /* remove possible idnet lookup node */
  if (!t->ident[0]) {
    ident_remove(t->ident);
  }

  while(t->data) {
    data_close(t->data);
  }

  free(t->buffer);
  free(t->obuffer);

  close(t->socket);

  if (t->status & ST_REMOTE)
    close(t->altsocket);

  printf("login_exit(%d/%d)\n", t->socket, t->altsocket);

  /* Remove from list and free() */
  for (curr = login_node_head; curr; last = curr, curr=curr->next) {
    if (curr == t) {
      if (!last)
        login_node_head = curr->next;
      else
        last->next = curr->next;

      free(t);
      break;

    }
  }

  if (mode_inetd)  /* from inetd, we quit when user logs out */
    exit(0);

#ifdef WIN32
  socket_restart();
#endif
}


void login_ident_reply(char *reply)
{
  struct login_node *run;
  /* find the logged in node, who's ident field == reply above.
   * check if it's pause, if so, unpause it and call _pass */

  for (run = login_node_head; run; run = run->next) {

    if ((char *)run->ident == reply) { /* found it */

      /* here, send ident reply to server, if we are connected. */
      if (run->status & ST_REMOTE) {

#ifdef SENDAUTH
	if (mode_auth) {

	  printf("Sending: IDEN %s@%s\n", run->ident, ul2a(run->host));

	  socket_print(run->altsocket, "IDEN %s@%s\r\n",
		       run->ident, ul2a(run->host));

	}
#endif

      } else { /* damn, no remote connect yet. Signal we have ident. */

	run->status |= ST_IDENT;

      }
    }
  }
}


void login_get(struct login_node *t, int in, int out)
{
  static unsigned char buffer[DATA_BUFSZ];
  char *buf_ptr;
  int red, wrote;

  buf_ptr = buffer;

  if ((red = read(in, buffer, DATA_BUFSZ)) <= 0) {

    /* Connection closed */
    t->status |= ST_LOGOUT;
    return;
  
  }

  /* parse input - skip sending if it found stuff */
  
  if (!mode_dumb && login_parse(t, buffer, red, in, out))
    return;

  /* send input */

  if ((wrote = write(out, buffer, red)) != red) {
    /* something went wrong */

    if (wrote < 0) {
      t->status |= ST_LOGOUT;
      return;
    }  

  }

}


void login_exit_check(void)
{
  struct login_node *runner, *next = NULL;

  for (runner = login_node_head; runner; runner = next) {

    next = runner->next;

    if (runner->status & ST_LOGOUT) {
      login_exit(runner);

    }

  }

  for (runner = login_node_head; runner; runner = next) {

    if (runner->data)
      data_exit_check((struct data_node *)runner->data);

  }

}

int login_parse(struct login_node *t, unsigned char *buffer, int size, 
		int in, int out)
{

  /* Right, check buffer for commands we need to roger. 
   * 
   * Those commands being:
   * PORT
   * PASV
   *
   */

  unsigned char *scan;

  if (!strncasecmp("PORT", buffer, 4)) {

    if (t->data) {
      data_close(t->data);
    }

    data_port(t, buffer, size, in, out);

    return 1;

  } else if (!strncasecmp("227 ", buffer, 4)) {

    if (t->data) {
      data_close(t->data);
    }

    data_pasv(t, buffer, size, in, out);

    return 1;
  }

  scan = buffer;

  return 0;
}


