#ifndef WIN32 
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif
#include <stdio.h>

#ifdef POSIX
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#endif
#include <signal.h>

#include "config.h"
#include "socket.h"
#include "io.h"


unsigned int mode_dumb     = 0;
unsigned int mode_portonly = 0;
unsigned int mode_pasvonly = 0;
unsigned int mode_auth     = 0;
unsigned int mode_remoteport   = 21;
unsigned int mode_localport    = 6699;
unsigned int mode_passiveport    = 0;
unsigned char *mode_remotehost = NULL;
unsigned int mode_daemon = 0;
unsigned int mode_inetd = 0;
unsigned int mode_useprint = 1;
unsigned int mode_usleep = 0;


extern int getopt();
extern char *optarg;


extern int io_exit;


void display_usage(char *app)
{

  printf("%s [-adhspliot]\n", app);
  printf("\t-a   : send IDEN string on connect (NOT RFC!) Use with LundFTPD\n");
  printf("\t-d   : dumb mode, do not relay data-sessions\n");
  printf("\t-h   : this text\n");
  printf("\t-s   : convert PORT data-sessions into PASV\n");
  printf("\t-p   : convert PASV data-sessions into PORT\n");
  printf("\t-i H : remote host to 'shadow'\n"); 
  printf("\t-o P : remote port to 'shadow' default 21\n");
  printf("\t-l P : local port to accept incomming connections\n");
  printf("\t-t P : when using PASV, use only port P (only 1 connection possible)\n");
  printf("\t-b   : detach and become a daemon\n");
  printf("\t-v   : inetd mode, use if this is to be started from inetd\n");
  printf("\t-u u : usleep for 'u' milliseconds in filecopy. (poor man's bandwidth limit)\r\n"); 

  
  exit(0);
}


char *mystrcpy(char *s)
{
  char *r;

  r = (char *) malloc(strlen(s)+1);
  if (!r) {
    perror("mystrcpy()");
    exit(1);
  }

  strcpy(r, s);
  return r;
}


void args_load(int argc, char **argv)
{
  int opt;

  while ((opt=getopt(argc, argv, "bvhdspai:o:t:l:u:")) != -1) {
    switch(tolower((char)opt)) {

    case 'h':
      display_usage(argv[0]);

    case 'd':
      mode_dumb ^= 1;
      break;

    case 's':
      mode_pasvonly ^= 1;
      break;

    case 'p':
      mode_portonly ^= 1;
      break;

    case 'a':
      mode_auth ^= 1;
      break;

    case 'o':
      mode_remoteport = atoi(optarg);
      break;

    case 'i':
      mode_remotehost = mystrcpy(optarg);
      break;

    case 'l':
      mode_localport = atoi(optarg);
      break;

    case 't':
      mode_passiveport = atoi(optarg);
      break;

    case 'b':
      mode_daemon ^= 1;
      break;

    case 'v':
      mode_inetd ^= 1;
      break;

    case 'u':
      mode_usleep = atoi(optarg);
      break;

    default:
      printf("What exactly is %c meant to do?\n", opt);
      display_usage(argv[0]);
      
    }
  }
}


void sig_int()
{
  signal(SIGINT, sig_int);
  io_exit = 1;
}




int main(int argc, char **argv)
{
#ifdef POSIX
  struct rlimit rlim;
#endif

  args_load(argc, argv);

  if (mode_inetd) {
    
    mode_useprint = 0;
    
    login_new(fileno(stdin));
    
  }
  
  printf("ShadowFTPD Version %s. (c) Jorgen Lundman <lundman@argonaut.com>\n", VERSION);
  

#ifndef WIN32
  if (mode_daemon) {
    
    close(fileno(stdin));
    close(fileno(stdout));
    close(fileno(stderr));
    
    setsid();
    
    if (fork()) exit(0);  /* Only child continues */
    
    mode_useprint = 0;

  } 
#endif



#ifdef POSIX

  getrlimit(RLIMIT_NOFILE, &rlim);

  rlim.rlim_cur = 256;

  setrlimit(RLIMIT_NOFILE, &rlim);

  printf("Max open file's: %d\n", getdtablesize());


  signal(SIGPIPE, SIG_IGN);


#endif





  socket_load();

  printf("Host %s:%u (%s) primed for connections.\n", 
	 mode_remotehost ? mode_remotehost : SHADOWHOST,
	 mode_remoteport, ul2a(shadowhost));


  signal(SIGINT, sig_int);


  io();


  socket_unload();

  return 0;
}


