#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#ifndef WIN32 
#include <unistd.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#else
#include <winsock.h>
#endif
#include <sys/types.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#ifndef INADDR_NONE
#define INADDR_NONE -1
#endif




#include "config.h"
#include "socket.h"


#ifdef WIN32
int myread(int s, char *buf, int len)
{
  return recv(s, buf, len, 0);
}
int mywrite(int s, char *buf, int len)
{
  return send(s, buf, len, 0);
}

#endif



unsigned long shadowhost = 0;
unsigned long localhost = 0;
int server_port = 0;

int socket_accept_fd = -1;
char localhost_a[80];


extern unsigned int mode_remoteport;
extern unsigned int mode_localport;
extern unsigned char *mode_remotehost;
extern unsigned char *mode_inetd;


char *ul2a(unsigned long q)
{
  static char host[64];

  sprintf(host, "%u.%u.%u.%u",
	  ((unsigned char *)&q)[0], /* Assume network order */
	  ((unsigned char *)&q)[1],
	  ((unsigned char *)&q)[2],
	  ((unsigned char *)&q)[3]);

  return host;
}


int socket_make(int *port)
     /* bind socket at port, if port = 0 it picks first free and sets it */
     /* returns -1 on error, or socket*/
{
  int c, sock;
  struct sockaddr_in l_sin;

  /* Here, open a socket ready for accept() */
#ifdef WIN32
  if ((sock = socket( AF_INET, SOCK_STREAM, 0 )) == INVALID_SOCKET) {
#else
  if ((sock = socket( AF_INET, SOCK_STREAM, 0 )) < 0 ) {
#endif
    perror("socket: ");
    return -1;
  }
  
  c = 1;
  setsockopt(sock,SOL_SOCKET ,SO_REUSEADDR,&c,sizeof(c));

#ifdef WIN32
  ioctlsocket(sock, FIONBIO, &c);
#else
  fcntl(sock,F_SETFL,(fcntl(sock,F_GETFL)|O_NONBLOCK));
#endif


  l_sin.sin_family = AF_INET;

#ifdef WIN32
  l_sin.sin_addr.s_addr = inet_addr("193.128.238.146");
//  l_sin.sin_addr.s_addr = htonl(inet_addr("0.0.0.0"));
#else
  l_sin.sin_addr.s_addr = htonl(INADDR_ANY);
#endif

  l_sin.sin_port = htons(*port); /* Any port */


  printf("bind()ing %s\n", ul2a(l_sin.sin_addr.s_addr));


#ifdef WIN32
  if ( bind( sock, (struct sockaddr *) &l_sin, sizeof(l_sin) ) == SOCKET_ERROR ) {
	errno = WSAGetLastError();
	printf("WSAGetLastError(): %d %x\n", errno, errno);
#else
  if ( bind( sock, (struct sockaddr *) &l_sin, sizeof(l_sin) ) ) {
#endif
    perror("bind: ");
    close(sock);
    return -1;
  }

  c = sizeof(struct sockaddr_in);
  getsockname(sock, (struct sockaddr *) &l_sin, &c);

  listen(sock, 5);

  *port = ntohs(l_sin.sin_port);
  return sock;
}


void socket_getIP(unsigned long *host, char *hostname)
{
  struct hostent *h;

  printf("Looking up '%s'...\n", hostname);
  
  if (!(h = gethostbyname(hostname))) {
    fprintf(stderr, "Can't determine hostname %s\r\n", hostname);
    exit(1);
  }
  
  bcopy(h->h_addr, host, sizeof(unsigned long));

}


void socket_load()
{


#ifdef WIN32

  WORD wVersionRequested;
  WSADATA wsaData;
  int err; 

  wVersionRequested = MAKEWORD( 2, 2 ); 
  err = WSAStartup( wVersionRequested, &wsaData );
  if ( err != 0 ) {
	printf("Couldn't start WinSock :( \n");
	exit(1);
  }

  printf("WinSock version %d.%d\n", LOBYTE( wsaData.wVersion ), HIBYTE( wsaData.wVersion ));

#endif



  if (!localhost) {

#ifdef WIN32
    if (gethostname(localhost_a, 80) != SOCKET_ERROR ) {
#else
    if (!gethostname(localhost_a, 80)) {
#endif
      socket_getIP(&localhost, localhost_a);

      printf("localhost is %s\n", ul2a(localhost));
    }
  }




  if (mode_remotehost)
    socket_getIP(&shadowhost, mode_remotehost);
  else
    socket_getIP(&shadowhost, SHADOWHOST);





  if (!mode_inetd) {

    server_port = mode_localport;
    
    if ((socket_accept_fd = socket_make(&server_port)) == -1) 
      exit(1);
    
    printf("port %d open for connections.\n", server_port);

  }

}


void socket_unload(void)
{

  printf("Freeing sockets\n");

  if (server_port) {
    close(server_port);
    server_port = 0;
  }

#ifdef WIN32
  WSACleanup( );
#endif
}


int socket_accept(int sock, unsigned long *remhost, int *remport)
{
  int new_sock;
  struct sockaddr_in from;
  int len = sizeof(from);

  new_sock = accept(sock, (struct sockaddr *)&from, &len);

#ifdef WIN32
  if (new_sock == INVALID_SOCKET) return -1;
#else
  if (new_sock < 0) return -1;
#endif

  bcopy(&from.sin_addr.s_addr, remhost, sizeof(unsigned long));
  *remport = ntohs(from.sin_port);


#ifndef LINUX

  /*
   *
   * This is by no means crucial for the ftpd to work, but will make it run
   * smoother. It is part of POSIX but lame OS, like Linux, decide only to
   * implement POSIX half-heartedly. But as always, the popular OS' are never
   * the good OS'. (See ms-dos, Win95, Linux etc)
   *
   */

  len = DATA_BUFSZ;
  setsockopt(new_sock, SOL_SOCKET, SO_SNDLOWAT, &len, sizeof(len));

#endif

#ifdef WIN32
  len = 1;
  ioctlsocket(sock, FIONBIO, &len);
#else
  fcntl(new_sock,F_SETFL,(fcntl(new_sock,F_GETFL)|O_NONBLOCK));
#endif

  return new_sock;
}


int socket_connect(unsigned long host, unsigned int port)
{
  struct sockaddr_in ssin;
  int sock, ret;

  bcopy(&host, &ssin.sin_addr, sizeof(host)) ;

#ifdef WIN32
  if ((sock = socket( AF_INET, SOCK_STREAM, 0 )) == INVALID_SOCKET) {
#else
  if ((sock = socket( AF_INET, SOCK_STREAM, 0 )) < 0) {
#endif
    perror("socket_connect(socket())");
    return -1;
  }

  ssin.sin_port = htons(port);
  ssin.sin_family = AF_INET;

#ifndef LINUX

  ret = DATA_BUFSZ;
  setsockopt(sock, SOL_SOCKET, SO_SNDLOWAT, &ret, sizeof(ret));

#endif

  /* Lets kick into NONBLOCKING */
#ifdef WIN32
  ret = 1;
  ioctlsocket(sock, FIONBIO, &ret);
#else
  fcntl(sock,F_SETFL,(fcntl(sock,F_GETFL)|O_NONBLOCK));
#endif

  /* Now, fire off the connect - make sure we didn't get any errors */
  ret = connect(sock,(struct sockaddr *)&ssin, sizeof(ssin));


#ifdef WIN32
  errno = WSAGetLastError();

  if((ret==SOCKET_ERROR) && errno!=WSAEINPROGRESS  && errno!=WSAEWOULDBLOCK) {
#else
  if(ret && errno!=EINPROGRESS && errno!=EWOULDBLOCK) {
#endif



    /* Doh, failed. */
    close(sock);
    return -1;
  }

  return sock;

}




#undef printf

#if __STDC__
void socket_print(int fd, char const *fmt, ...)
#else
void socket_print(fd, fmt, va_alist)
     int fd;
     char *fmt;
     va_dcl
#endif
{
  va_list ap;
  char msg[1024];

#if __STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif

  vsprintf(msg, fmt, ap);
  va_end(ap);

  write(fd, msg, strlen(msg));
}





#ifdef WIN32
void socket_restart(void) 
{
  int c, sock = socket_accept_fd;
  struct sockaddr_in l_sin;

#if 0
  l_sin.sin_family = AF_INET;

  l_sin.sin_addr.s_addr = inet_addr("193.128.238.146");

  l_sin.sin_port = htons(6699); /* Any port */


  printf("restart bind()ing %s\n", ul2a(ntohl(l_sin.sin_addr.s_addr)));


  if ( bind( sock, (struct sockaddr *) &l_sin, sizeof(l_sin) ) == SOCKET_ERROR ) {
	errno = WSAGetLastError();
	printf("error %d %x\n", errno, errno);
    perror("bind: ");
    close(sock);
    return -1;
  }


  if (listen(socket_accept_fd, 5) == SOCKET_ERROR)
	printf("Listen failed! %d\n", WSAGetLastError());
  
  printf("port %d open for connections.\n", server_port);

#endif

  close(socket_accept_fd);


  while (1) {

	SleepEx(5000, FALSE);

	
	if ((socket_accept_fd = socket_make(&server_port)) != -1) 
	  break;

  }
	  
  
  
}
#endif
